
from enum import Enum
from pydantic import BaseModel



class ErrorTypes(Enum):
    AgencyError: int = 0
    AgencyErrorDuplicateID: int = 1
    AgencyErrorInvalidAgency: int = 2

    GTFSError: int = 100
    GTFSErrorInvalidFile: int = 101
    GTFSErrorUnknownFile: int = 102
    GTFSErrorProcessReturnedFalse: int = 103

    StopError: int = 200
    StopErrorDuplicateID: int = 201
    StopErrorInvalidStop: int = 202



class AgencyError(BaseModel):
    caller: str = "Agency"
    name: str = "AgencyError"
    error_type: ErrorTypes = ErrorTypes.AgencyError
    message: str
    values: list = []



class GTFSError(BaseModel):
    caller: str = "GTFS"
    name: str = "GTFSError"
    error_type: ErrorTypes = ErrorTypes.GTFSError
    message: str
    values: list = []



class StopError(BaseModel):
    caller: str = "Stop"
    name: str = "StopError"
    errory_type: ErrorTypes = ErrorTypes.StopError
    message: str
    values: list = []



from pydantic import BaseModel, validator
from typing import Optional, Union
from ..models.models import (
    LocationTypes,
    WheelchairBoarding_Parentless,
    WheelchairBoarding_Child,
    WheelchairBoarding_Entrance,
    WheelchairBoarding_Exit
)



class Stop(BaseModel):
    """
        https://developers.google.com/transit/gtfs/reference#stopstxt

        :param stop_id [str] -- Identifies a stop, station, or station entrance (station entrance refers to both entrances and exits) (multiple routes may use the same stop)
        :param stop_code [Optional[str]] -- Short text or a number that identifies the location for the riders
        :param stop_name [Optional[str]] -- Name of the stop that local people and tourists understand (required if location_type = 0, 1 or 2 )
        :param stop_desc [Optional[str]] -- Description of the location, that provides useful information
        :param stop_lat [Optional[float]] -- Latitude of the location (required if location_type = 0, 1 or 2)
        :param stop_lon [Optional[float]] -- Longitude of the location (required if location_type = 0, 1 or 2)
        :param zone_id [Optional[str]] -- The fare zone for the stop (required if providing fare information in 'fare_rules.txt' )
        :param stop_url [Optional[str]] -- URL of a web page about the location (should not be the same as the agency url and the route_url)
        :param location_type [Optional[LocationTypes]] -- The type of location (see 'models.models.LocationTypes' for details)
        :param parent_station [Optional[str]] -- ID of the parent location (required if location_type = 2, 3 or 4; optional if location_type = 0; forbidden if location_type = 1)
        :param stop_timezone [Optional[str]] -- The timezone of the location (if it has a parent station, it will inherit from the parent station)
        :param wheelchair_boarding [Optional[WheelchairBoarding]] -- Indicates if wheelchair boarding is possible from this stop (one of the following (models.models): [WheelchairBoardin_Parentless, WheelchairBoarding_Child, WheelchairBoarding_Entrance, WheelchairBoarding_Exit])
        :param level_id [Optional[str]] -- The level of the location (multiple stations can have the same level)
        :param platform_code [Optional[str]] -- Identifier for a platform stop (ex.: "A1" )
    """
    stop_id: str
    stop_code: Optional[str]
    stop_name: Optional[str]
    stop_desc: Optional[str]
    stop_lat: Optional[float]
    stop_lon: Optional[float]
    zone_id: Optional[str] # required if providing fare information using 'fare_rules.txt'
    stop_url: Optional[str]
    location_type: Optional[LocationTypes]
    parent_station: Optional[str]
    stop_timezone: Optional[str]
    wheelchair_boarding: Optional[Union[
        WheelchairBoarding_Parentless,
        WheelchairBoarding_Child,
        WheelchairBoarding_Entrance,
        WheelchairBoarding_Exit]
        ]
    level_id: Optional[str]
    platform_code: Optional[str]



    @validator("stop_name", always=True)
    def is_name_valid(cls, name, values):
        """
            The field 'stop_name' is required if the 'location_type' is one of the 'matching_types'
        """
        matching_types: list = [
            LocationTypes.STOP,
            LocationTypes.PLAFTORM,
            LocationTypes.STATION,
            LocationTypes.ENTRANCE,
            LocationTypes.EXIT
            ]
        
        if values["location_type"] != None:
            if values["location_type"] in matching_types:
                if name == None or len(name) < 1:
                    raise ValueError(f"is required if 'location_type' matches one of the following: \n {matching_types}")

        return name
    

    @validator("stop_lat", always=True)
    def is_lat_valid(cls, lat, values):
        """The field 'stop_lat' is required if the 'location_type' is one of the 'matching_types'"""
        matching_types: list = [
            LocationTypes.STOP,
            LocationTypes.PLAFTORM,
            LocationTypes.STATION,
            LocationTypes.ENTRANCE,
            LocationTypes.EXIT
        ]

        if values["location_type"] != None and values["location_type"] in matching_types:
            if lat == None:
                raise ValueError(f"is required if 'location_type' matches one of the following: \n {matching_types}")
            
        return lat
    

    @validator("stop_lon", always=True)
    def is_lat_valid(cls, lon, values):
        """The field 'stop_lon' is required if the 'location_type' is one of the 'matching_types'"""
        matching_types: list = [
            LocationTypes.STOP,
            LocationTypes.PLAFTORM,
            LocationTypes.STATION,
            LocationTypes.ENTRANCE,
            LocationTypes.EXIT
        ]

        if values["location_type"] != None and values["location_type"] in matching_types:
            if lon == None:
                raise ValueError(f"is required if 'location_type' matches one of the following: \n {matching_types}")
            
        return lon
    

    @validator("parent_station", always=True)
    def is_parent_station_valid(cls, parent_station, values):
        """
            Required:
                - Entrances/Exits (location_type=2)
                - Generic Nodes (location_type=3)
                - Boarding Areas (location_type=4)
            
            Forbidden:
                - Stations (location_type=1)

            Optional:
                - Stops/Platforms (location_type=0 or empty)
        """

        required: list = [
            LocationTypes.ENTRANCE,
            LocationTypes.EXIT,
            LocationTypes.GENERIC_NODE,
            LocationTypes.BOARDING_AREA
        ]

        forbidden: list = [
            LocationTypes.STATION
        ]

        if values["location_type"] != None:
            if values["location_type"] in required and parent_station == None or len(parent_station) < 1:
                raise ValueError(f"required if 'location_type' matches any of the following: \n {required}")
            
            elif values["location_type"] in forbidden and parent_station != None:
                raise ValueError(f"forbidden if 'location_type' matches any of the following: \n {forbidden}")
            
        return parent_station



class ErrorManager():
    def __init__(self, silent_errors: bool = False) -> None:
        self.silent_errors: bool = silent_errors
        self._errors: list = []

    
    def set_silent_errors(self, silent_errors: bool) -> bool:
        if isinstance(silent_errors, bool):
            self.silent_errors = silent_errors
            return True
        
        return False
    
    
    def get_silent_errors(self) -> bool:
        return self.silent_errors
    
    
    def is_silent_errors(self) -> bool:
        return self.silent_errors
    

    def errors(self) -> list:
        return self._errors


    def clear(self) -> bool:
        self._errors = []
        return True

    
    def add_error(self, error) -> bool:
        self._errors.append(error)
        return True
    


"""
    Some utility functions used all around the files
"""

from . import variables
from .managers.errormanager import ErrorManager



def get_error_manager() -> ErrorManager:
    """
        Returns the current error manager, if none is set it will create one.

        :return [ErrorManager]
    """

    if variables.ERRORMANAGER == None:
        return init_error_manager()
    
    return variables.ERRORMANAGER



def init_error_manager(silent_errors: bool = False) -> ErrorManager:
    """
        Creates a new error manager.

        :param silent_errors [bool] -- Whether or not it should raise errors which are not critical

        :return [ErrorManager]
    """

    new_error_manager = ErrorManager(silent_errors=silent_errors)
    new_error_manager.clear()

    return new_error_manager



def set_error_manager(error_manager: ErrorManager) -> bool:
    """
        Sets the package error manager to the manager you passed.

        :param error_manager [ErrorManager] -- An ErrorManager class object

        :return [bool] -- True on success, False on fail
    """

    if isinstance(error_manager, ErrorManager):
        variables.ERRORMANAGER = error_manager
        return True
    
    return False






ERRORMANAGER = None


REQUIRED_OPTIONS: dict = {
    "always": "This file always has to be part of the dataset.",
    "conditional": "This file has to be part of the dataset only in certain cases.",
    "optional": "This file does not need to be part of the dataset."
}


FILES: dict = {
    "agency.txt": {
        "required": "always"
    },
    "stops.txt": {
        "required": "always"
    },
    "routes.txt": {
        "required": "always"
    },
    "trips.txt": {
        "required": "always"
    },
    "stop_times.txt": {
        "required": "always"
    },
    "calendar.txt": {
        "required": "conditional"
    },
    "calendar_dates.txt": {
        "required": "conditional"
    },
    "fare_attributes.txt": {
        "required": "optional"
    },
    "fare_rules.txt": {
        "required": "optional"
    },
    "shapes.txt": {
        "required": "optional"
    },
    "frequencies.txt": {
        "required": "optional"
    },
    "transfers.txt": {
        "required": "optional"
    },
    "pathways.txt": {
        "required": "optional"
    },
    "levels.txt": {
        "required": "optional"
    },
    "feed_info.txt": {
        "required": "conditional"
    },
    "translations.txt": {
        "required": "optional"
    },
    "attributions.txt": {
        "required": "optional"
    }
}


OUTPUT_FOLDER: str = "/data"


"""
    Module to manage multiple 'Agency' models, write them to a file, load them from a file or parse them from a json object.

    Author: Felix Michelis
    Date: 12/05/2023
        dd/mm/yyyy
"""
from pydantic import ValidationError
from typing import Optional, List
import csv
from collections import Counter
import json


from ..utility import get_error_manager
from ..objects.agency import Agency
from ..models.errors import ErrorTypes, AgencyError




class Agencies():
    """A class to manage multiple 'Agency' models and write them to a file, load them from a file or parse them from a json object."""

    def __init__(self, agencies: List[Agency] = None, file: str = None):
        """
            A class to manage multiple 'Agency' models and write them to a file, load them from a file or parse them from a json object.

            :param agencies [List[Agency]] -- A list of 'Agency' models, can be passed later with the respective function on function call
            :param file [str] -- The file to write to or load from, can be passed later with the respective function on function call
        """
        self.error_manager = get_error_manager()
        self.agencies: List[Agency] = agencies
        self.file: str = file
    

    def to_file(self, agencies: Optional[List[Agency]] = None, file: Optional[str] = None) -> bool:
        """
            Writes a list of 'Agency' models to a file in csv format.

            :param agencies [Optional[List[Agency]]] -- A list containing 'Agency' models, if not given it will take 'self.agencies' if set
            :param file [Optional[str]] -- The file to write to, if not given it will take 'self.file' if set

            :return bool -- True on success, else False or raises Error if not silent error enabled
        """

        if not agencies:
            if self.agencies == None:
                raise TypeError("Got no agencies to write to file!")
            agencies = self.agencies

        if not file:
            if self.file == None:
                raise TypeError("Got no file to write to!")
            file = self.file


        headers: list = [variable for variable in agencies[0].dict(exclude_unset=False).keys()] # create a list of the headers for the csv file
        rows: list = []
        
        agencies_ids = [] # this list stores the ids of all the agencies that will be written to the file so we can verify that no agencies have the same id
        for obj in agencies:
            obj_as_dict = obj.dict(exclude_unset=False)
            agencies_ids.append(obj_as_dict["agency_id"]) # add the id of the agency to our ids list for verification later on
            rows.append(obj_as_dict) # add the agency as json to our rows that will be written to the csv file

        if len(agencies_ids) > 1:
            duplicates = [k for k, v in Counter(agencies_ids).items() if v>1] # returns a list with all duplicate ids; if the id '123' was for example at least 2 times in the ids list the output would be: ['123']
            if len(duplicates) > 0:
                if self.error_manager.is_silent_errors():
                    self.error_manager.add_error(AgencyError(
                        caller = "AgencyManager",
                        error_type=ErrorTypes.AgencyErrorDuplicateID,
                        message="Multiple agencies cannot have the same id!",
                        values = duplicates
                    ))
                    return False
                else:
                    raise ValueError(f"Multiple agencies cannot have the same id! Got the following duplicate ids: {duplicates}")
            

        with open(file, "w", newline='') as _file:
            csvwriter = csv.DictWriter(_file, fieldnames=headers)
            csvwriter.writeheader()
            csvwriter.writerows(rows)


        return True



    def parse(self, agencies_list: List[dict]) -> List[Agency]:
        """
            Converts a list of dictionaries which each represent an 'Agency' to a list of 'Agency' models.
        """
        agencies = []

        for agency in agencies_list:
            if self.is_valid_agency(agency):
                agencies.append(Agency.parse_obj(agency))
            else:
                if self.error_manager.is_silent_errors():
                    self.error_manager.add_error(AgencyError(
                        caller = "AgencyManager",
                        error_type = ErrorTypes.AgencyErrorInvalidAgency,
                        message="Invalid agency from dict!",
                        values = [agency]
                    ))
                    return False
                else:
                    raise ValueError(f"Invalid agency: {agency}")

        return agencies
    


    def load_file(self, file: Optional[str] = None) -> List[Agency]:
        """
            Reads a file in csv format and outputs a list of 'Agency' models contained in that file.

            :param file [Optional[str]] -- The file to read from, if not given it will take 'self.file' if its set

            :return List[Agency] -- A list conatining 'Agency' models
        """

        if not file:
            if self.file == None:
                raise TypeError("Got no file to load from!")
            file = self.file
        

        with open(file, "r") as _file:
            csvreader = list(csv.DictReader(_file)) # gives a list of dictionaries where each dict is a json representation of an agency

        agencies = self.parse(csvreader)

        agencies_ids = []

        for agency in agencies:
            if agency.agency_id in agencies_ids:
                if self.error_manager.is_silent_errors():
                    self.error_manager.add_error(AgencyError(
                        caller = "AgencyManager",
                        error_type = ErrorTypes.AgencyErrorDuplicateID,
                        message="Cannot have multiple agencies with the same id!",
                        values = [agency.agency_id]
                    ))
                    return False
                else:
                    raise ValueError("Cannot have multiple agencies with the same id!")
            agencies_ids.append(agency.agency_id)

        return agencies

        
    def is_valid_agency(self, agency: dict) -> bool:
        """
            Checks if a dict is a valid 'Agency' model.

            :param agency [dict] -- The dict to validate
            
            :return bool -- 'True' if it contains a valid model, otherwise ('False', dict[argument_causing_error: error_message])
        """

        try:
            Agency.parse_obj(agency)

        except ValidationError as e:
            errors = {}

            for error in json.loads(e.json()):
                errors[error["loc"][0]] = error["msg"]

            return (False, errors)
        
        return True


"""
    This class handles all the other managers (altough they each work on their own).
    You can use this class to quickly read a gtfs dataset or create one from your data without calling each single manager by hand.
"""

import os
from zipfile import ZipFile
import csv
import io
import codecs


from ..utility import get_error_manager
from ..variables import FILES, OUTPUT_FOLDER
from .agencymanager import Agencies
from .stopmanager import Stops
from ..models.errors import ErrorTypes, GTFSError



class GTFSManager():
    def __init__(self, silent_errors: bool = False) -> None:
        self.silent_errors = silent_errors
        self.error_manager = get_error_manager()
        self.error_manager.set_silent_errors(self.silent_errors)
        
            

    def load_folder(self, input_folder: str, is_zip: bool = True, output_folder: str = None) -> dict:
        """
            Loads a folder adhering to the gtfs standards.

            :param input_folder [str] -- The path to the input folder
            :param is_zip [Optional[bool]] -- If the provided folder is a zip folder or not (default=True)
            :param output_folder [Optional[str]] -- The path to the folder into which the files will be extracted if you provide a zip folder

        """

        data: dict = {}
        
        if is_zip:
            #utf8_reader = codecs.getreader("utf-8")
            
            with ZipFile(input_folder, "r") as folder:
                if output_folder == None:
                    output_folder = str(str(os.getcwd()) + OUTPUT_FOLDER + "/" + str(".".join(input_folder.split(".")[:-1]))) # Returns the name of the folder without the ending (could also do input_folder.replace(".zip", "") )
                    
                folder.extractall(path = output_folder)
                
            
            self.load_folder(input_folder = output_folder, is_zip=False)
                #for file in folder.namelist():
                    #with folder.open(file, "r") as _file:
                        #a = utf8_reader(_file) or:
                        #a = io.StringIO(_file.read().decode())
                        #data[file] = list(csv.DictReader(a))

                        #self.parse_DictReader(data[file])

        if not is_zip:
            data["folder"] = str(input_folder)

            for file in list(os.listdir(input_folder)):
                filepath = str(input_folder + "/" + file)
                file_lower = str(file).lower()

                if file_lower == "agency.txt":
                    ag = Agencies()
                    data["agency"] = ag.load_file(filepath) 

                elif file_lower == "stops.txt":
                    st = Stops()
                    data["stops"] = st.load_file(filepath)

                else:
                    if self.error_manager.is_silent_errors():
                        self.error_manager.add_error(GTFSError(
                            error_type=ErrorTypes.GTFSErrorUnknownFile,
                            message="Unknown file in dataset.",
                            values = [filepath]
                        ))
                    else:
                        raise ValueError(f"Unknown file in dataset: {filepath}")


        return data
    

    def save_to_folder(self, data: dict, output_folder: str, skip_unknown: True) -> bool:
        """
            Saves data adhering to the gtfs standards to a folder.

            :param data [dict] -- The gtfs data to write to the folder
            :param output_folder [Optional[str]] -- The path to the output folder (it will not create a new folder in the folder you pass)
            :param skip_unknown [Optional[bool]] -- Skip unknown keys

            :return [bool] -- True on success, False on fail
        """

        for _key in data:
            key_lower = str(_key).lower()

            ## Agencies
            if key_lower == "agency":
                file = str(output_folder + "/agency.txt")
                cl = Agencies()
                success = cl.to_file(data[_key], file=file)

                if not success:
                    if self.error_manager.is_silent_errors():
                        self.error_manager.add_error(GTFSError(
                            error_type=ErrorTypes.GTFSErrorProcessReturnedFalse,
                            message = "'Agencies().to_file()' returned False upon execution. (Maybe check previous errors?)",
                            values = [file, data[_key]]
                        ))
                    else:
                        raise InterruptedError(f"Process interrupted while saving file: {file} \n Function 'Agencies().to_file()' did not return True upon execution. (Check silent errors log?) \n Data: {data[_key]}")
                    
            ## Stops
            elif key_lower == "stops":
                file = str(output_folder + "/stops.txt")
                cl = Stops()
                success = cl.to_file(data[_key], file=file)

                if not success:
                    if self.error_manager.is_silent_errors():
                        self.error_manager.add_error(GTFSError(
                            error_type=ErrorTypes.GTFSErrorProcessReturnedFalse,
                            message = "'Stops().to_file()' returned False upon execution. (Maybe check previous errors?)",
                            values = [file, data[_key]]
                        ))
                    else:
                        raise InterruptedError(f"Process interrupted while saving file: {file} \n Function 'Stops().to_file()' did not return True upon execution. (Check silent errors log?) \n Data: {data[_key]}")
                




from pydantic import BaseModel, validator
from typing import Optional
from urllib.parse import urlparse


class Agency(BaseModel):
    """
        https://developers.google.com/transit/gtfs/reference#agencytxt

        :param id [Optional[str]] -- The unique id for this agency (only needed if multiple agencies are present in the dataset)
        :param name [str] -- The full name of the agency
        :param url [str] -- The url of the agency (f.ex.: 'https://example.com' )
        :param timezone [str] -- The timezone of the agency (if multiple agencies in this dataset, all must have the same timezone)
        :param lang [Optional[str]] -- The primary language this agency uses
        :param phone [Optional[str]] -- A phone number for the agency
        :param fare_url [Optional[]] -- The website on which travellers can buy tickets to travel with this agency
        :param email [Optional[str]] -- An email address to reach the agency (should be the support email address of the agency or another email which is actively used by the agency)
    """
    agency_id: Optional[str] = ""
    agency_name: str
    agency_url: str
    agency_timezone: str
    agency_lang: Optional[str] = ""
    agency_phone: Optional[str] = ""
    agency_fare_url: Optional[str] = ""
    agency_email: Optional[str] = ""


    @validator("agency_url")
    def is_agency_url_valid(cls, url):
        if isinstance(url, str):      
            # https://stackoverflow.com/a/52455972/21595907
            try:
                result = urlparse(url)
                if all([result.scheme, result.netloc]):
                    return url

            except ValueError:
                # ValueError can happen if an invalid ip (v4, v6) or a byte object has been passed
                return ValueError("must be a valid url (f.ex.: https://example.com)")
        
        raise ValueError("must be a valid url (f.ex.: https://example.com)")
    

    @validator("agency_fare_url")
    def is_agency_fare_url_valid(cls, url):
        if isinstance(url, str):
            if len(url) < 1: # url is empty and we can accept that (mainly for when parsing from file)
                return url
            
            # https://stackoverflow.com/a/52455972/21595907
            try:
                result = urlparse(url)
                if all([result.scheme, result.netloc]):
                    return url

            except ValueError:
                # ValueError can happen if an invalid ip (v4, v6) or a byte object has been passed
                return ValueError("must be a valid url (f.ex.: https://example.com)")
        
        raise ValueError("must be a valid url (f.ex.: https://example.com)")





from enum import Enum
from pydantic import BaseModel


class LocationTypes(Enum):
    """
        Describes the type of location (station, stop, boarding area, ...)

        :STOP = 0
        :PLATFORM = 0
        :STATION = 1
        :ENTRANCE = 2
        :EXIT = 2
        :GENERIC_NODE = 3
        :BOARDING_AREA = 4
    """
    STOP: int = 0
    PLAFTORM: int = 0
    STATION: int = 1
    ENTRANCE: int = 2
    EXIT: int = 2
    GENERIC_NODE: int = 3
    BOARDING_AREA: int = 4


class WheelchairBoarding_Parentless(Enum):
    """
        Wheelchair accessibility for stops without parents

        :EMPTY = 0 -- No accessibility information for the stop
        :UNKNOWN = 0 -- equal to 'EMPTY'
        :PARTIALLY = 1 -- Some vehicles at this stop can be boarded by a rider in a wheelchair
        :NO = 2 -- Some vehicles at this stop can be boarded by a rider in a wheelchair
        :IMPOSSIBLE = 2 -- equal to 'NO'
    """
    EMPTY: int = 0
    UNKNOWN: int = 0
    PARTIALLY: int = 1
    NO: int = 2
    IMPOSSIBLE: int = 2


class WheelchairBoarding_Child(Enum):
    """
        Wheelchair accessibility for stops that are children of a parent stop

        :EMPTY = 0 -- Stop will inherit its wheelchair_boarding behavior from the parent station, if specified in the parent
        :UNKNOWN = 0 -- equal to 'EMPTY'
        :YES = 1 -- There exists some accessible path from outside the station to the specific stop/platform
        :POSSIBLE = 1 -- equal to 'YES'
        :NO = 2 -- There exists no accessible path from outside the station to the specific stop/platform
        :IMPOSSIBLE = 2 -- equal to 'NO'
    """
    EMPTY: int = 0 # will inherit accessibility from parent
    UNKNOWN: int = 0 # will inherit accessibility from parent
    YES: int = 1
    POSSIBLE: int = 1
    NO: int = 2
    IMPOSSIBLE: int = 2


class WheelchairBoarding_Entrance(Enum):
    """
        Wheelchair accessibility for entrances

        :EMPTY = 0 -- Station entrance will inherit its wheelchair_boarding behavior from the parent station, if specified for the parent
        :UNKOWN = 0 -- equal to 'EMPTY'
        :YES = 1 -- Station entrance is wheelchair accessible
        :POSSIBLE = 1 -- equal to 'YES'
        :NO = 2 -- No accessible path from station entrance to stops/platforms
        :IMPOSSIBLE = 2 -- equal to 'NO'

    """
    EMPTY: int = 0 # will inherit accessibility from parent
    UNKOWN: int = 0 # will inherit accessibility from parent
    YES: int = 1
    POSSIBLE: int = 1
    NO: int = 2
    IMPOSSIBLE: int = 2


class WheelchairBoarding_Exit(Enum):
    """
        Wheelchair accessibility for exits (equal to entrances)

        :EMPTY = 0 -- Station exit will inherit its wheelchair_boarding behavior from the parent station, if specified for the parent
        :UNKOWN = 0 -- equal to 'EMPTY'
        :YES = 1 -- Station exit is wheelchair accessible
        :POSSIBLE = 1 -- equal to 'YES'
        :NO = 2 -- No accessible path from station stops/platforms to exit
        :IMPOSSIBLE = 2 -- equal to 'NO'

    """
    EMPTY: int = 0 # will inherit accessibility from parent
    UNKOWN: int = 0 # will inherit accessibility from parent
    YES: int = 1
    POSSIBLE: int = 1
    NO: int = 2
    IMPOSSIBLE: int = 2



class RouteType(Enum):
    TRAM: int = 0
    Streetcar: int = 0
    LIGHT_RAIL: int = 0
    STREET_LEVEL_SYSTEM: int = 0

    SUBWAY: int = 1
    METRO: int = 1
    UNDERGROUND_LEVEL_SYSTEM: int = 1

    RAIL: int = 2
    INTERCITY: int = 2
    LONG_DISTANCE: int = 2

    BUS: int = 3
    LONG_DISTANCE_BUS: int = 3
    SHORT_DISTANCE_BUS: int = 3

    FERRY: int = 4
    LONG_DISTANCE_FERRY: int = 4
    SHORT_DISTANCE_FERRY: int = 4

    CABLE_TRAM: int = 5

    AERIAL_LIFT: int = 6
    SUSPENDED_CABLE_CAR: int = 6
    CABLE_CAR: int = 6
    GONDOLA_LIFT: int = 6
    AERIAL_TRAMWAY: int = 6
    CABLE_TRANSPORT: int = 6

    FUNICULAR: int = 7

    TROLLEYBUS: int = 11
    
    MONORAIL: int = 12




class ContinuousPickup(Enum):
    """
        Indicates whether a rider can board the transit vehicle anywhere along the vehicle’s travel path. The path is described by shapes.txt on every trip of the route.

            :CONTINUOUS = 0 -- Continuous stopping pickup
            :NO = 1 or empty -- No continuous stopping pickup
            :NON_CONTINUOUS = 1 -- equal to 'NO'
            :PHONE = 2 -- Must phone an agency to arrange continuous stopping pickup
            :PHONE_AGENCY = 2 -- equal to 'PHONE'
            :DRIVER = 3 -- Must coordinate with a driver to arrange continuous stopping pickup
            :COORDINATE_WITH_DRIVER = 3 -- equal to 'DRIVER'
    """
    CONTINUOUS: int = 0
    NO: int = 1
    NON_CONTINUOUS: int = 1
    PHONE: int = 2
    PHONE_AGENCY: int = 2
    DRIVER: int = 3
    COORDINATE_WITH_DRIVER: int = 3


class ContinuousDropOff(Enum):
    """
        Indicates whether a rider can alight from the transit vehicle at any point along the vehicle’s travel path. The path is described by shapes.txt on every trip of the route.

            :CONTINUOUS = 0 -- Continuous stopping drop-off
            :NO = 1 or empty -- No continuous stopping drop-off
            :NON_CONTINUOUS = 1 -- equal to 'NO'
            :PHONE = 2 -- Must phone an agency to arrange continuous stopping drop-off
            :PHONE_AGENCY = 2 -- equal to 'PHONE'
            :DRIVER = 3 -- Must coordinate with a driver to arrange continuous stopping drop-off
            :COORDINATE_WITH_DRIVER = 3 -- equal to 'DRIVER'
    """
    CONTINUOUS: int = 0
    NO: int = 1
    NON_CONTINUOUS: int = 1
    PHONE: int = 2
    PHONE_AGENCY: int = 2
    DRIVER: int = 3
    COORDINATE_WITH_DRIVER: int = 3


class DirectionId(Enum):
    """
        Indicates the direction of travel for a trip.

        :OUTBOUND = 0 -- Travel in one direction (e.g. outbound travel)
        :INBOUND = 1 -- Travel in the opposite direction (e.g. inbound travel)
    """
    OUTBOUND: int = 0
    INBOUND: int = 1


class WheelchairAccessible(Enum):
    """
        Indicates wheelchair accessibility.

        :UNKNOWN = 0 -- No accessibility information for the trip
        :YES = 1 -- Vehicle being used on this particular trip can accommodate at least one rider in a wheelchair
        :NO = 2 -- No riders in wheelchairs can be accommodated on this trip
    """
    UNKNOWN: int = 0
    YES: int = 1
    NO: int = 2



class BikeAllowed(Enum):
    """
        Indicates whether bikes are allowed.

        :UNKNOWN = 0 -- No bike information for the trip
        :YES = 1 -- Vehicle being used on this particular trip can accommodate at least one bicycle
        :NO = 2 -- No bicycles are allowed on this trip
    """
    UNKNOWN: int = 0
    YES: int = 1
    NO: int = 2
